/*
 * $Source: /home/gjvc/rnd/apache/modules/mod_log_access/mod_log_access-CURRENT/RCS/mod_log_access.c,v $
 * $Id: mod_log_access.c,v 1.37 2007/02/04 10:22:09 gjvc Exp $
 */


/* UNIX */
#include <unistd.h>

/* APR */
#include "apr_version.h"
#include "apr_anylock.h"
#include "apr_hash.h"
#include "apr_lib.h"
#include "apr_optional.h"
#include "apr_strings.h"
#include "apr_pools.h"
#include "apr_file_io.h"
#include "util_time.h"

/* httpd */
#include "ap_mpm.h"
#include "ap_config.h"
#include "httpd.h"
#include "http_core.h"
#include "http_config.h"
#include "http_log.h"
#include "http_protocol.h"

/* log_timeutils */
#include "log_timeutils.h"


static const char *RCS_ID = "$Id: mod_log_access.c,v 1.37 2007/02/04 10:22:09 gjvc Exp $";


/* ---- begin constants ---- */

#define DEFAULT_REQUEST_TIME_SIZE   32                          /* */
#define TIME_CACHE_SIZE             4                           /* */
#define TIME_CACHE_MASK             3                           /* */

#define LOG_FORMAT_CLF              "%h %l %u %t \"%r\" %>s %b" /* */

#define LOG_FILE_FOPEN_FLAGS        ( APR_FOPEN_WRITE | APR_FOPEN_APPEND | APR_FOPEN_CREATE | APR_XTHREAD )
#define LOG_FILE_PERMISSIONS        ( APR_OS_DEFAULT )

/* ---- end constants ---- */


/* ---- begin types ---- */

#define LOG_TYPE_BUFFERED       ( 0001 )    /* */
#define LOG_TYPE_CONDITIONAL    ( 0002 )    /* */
#define LOG_TYPE_FILE           ( 0004 )    /* */
#define LOG_TYPE_PIPE           ( 0010 )    /* */
#define LOG_TYPE_ROLLING        ( 0020 )    /* */
#define LOG_TYPE_SYMLINKED      ( 0040 )    /* */
#define LOG_TYPE_ERROR_LOG      ( 0100 )    /* */
#define LOG_TYPE_ACCESS_LOG     ( 0200 )    /* */

#define LOG_IS_BUFFERED(x)      ( ((x) & LOG_TYPE_BUFFERED)    == LOG_TYPE_BUFFERED    )
#define LOG_IS_CONDITIONAL(x)   ( ((x) & LOG_TYPE_CONDITIONAL) == LOG_TYPE_CONDITIONAL )
#define LOG_IS_FILE(x)          ( ((x) & LOG_TYPE_FILE)        == LOG_TYPE_FILE        )
#define LOG_IS_PIPE(x)          ( ((x) & LOG_TYPE_PIPE)        == LOG_TYPE_PIPE        )
#define LOG_IS_ROLLING(x)       ( ((x) & LOG_TYPE_ROLLING)     == LOG_TYPE_ROLLING     )
#define LOG_IS_SYMLINKED(x)     ( ((x) & LOG_TYPE_SYMLINKED)   == LOG_TYPE_SYMLINKED   )
#define LOG_IS_ERROR_LOG(x)     ( ((x) & LOG_TYPE_ERROR_LOG)   == LOG_TYPE_ERROR_LOG   )
#define LOG_IS_ACCESS_LOG(x)    ( ((x) & LOG_TYPE_ACCESS_LOG)  == LOG_TYPE_ACCESS_LOG  )


/*

  server_context -- per-[virtual-]server configuration.
  
    log_contexts is an apr_array of log_context structures
  
  If a default log format is given by LogFormat, store in default_format_items.
  
  for the main server
    * parent_log_contexts is NULL
    * log_contexts points to a list of the log files
  
  in each virtual host
    * parent_log_contexts points to the same array as log_contexts in
      the main server, so that if this virtual host has no logs defined,
      default to the main server's logs.
    * log_contexts points to the array of logs defined for this virtual
      host, which might be empty. 

  server
    parent_log_contexts -> NULL
    log_contexts -> [ list of log files ]

  virtual host
    parent_log_contexts -> &server.log_contexts;
    log_contexts -> [ logs, ... ]

*/

typedef struct server_context {

    const char          *server_hostname;       /* */

    apr_table_t         *log_format_table;      /* <format-name, format-string> pairs registered using the LogFormat directive */

    const char          *default_format_string; /* default format string specified by LogFormat directive */
    apr_array_header_t  *default_format_items;  /* array of function pointers */

    apr_array_header_t  *parent_log_contexts;   /* ${if-this-is-the-main-server} ? NULL : pointer to parent server_context.log_contexts */
    apr_array_header_t  *log_contexts;          /* pointer to array of pointers to log_context structs (access logs) */
    apr_array_header_t  *error_log_contexts;    /* pointer to array of pointers to log_context structs (error logs) */

    int                 clf_timestamps;         /* if set, create timestamps according to CLF convention */

} server_context_t;


/*

  log_context -- status of a single log file

  if filename is NULL then
    no logging is done for this request

  if log_format_string is non-NULL then
    format = log_format_string
  else
    if server_context->default_format_items is non-NULL
        format = server_context->default_format_items
    else
        format = LOG_FORMAT_CLF
 
*/

typedef struct log_context {

    const char          *filename;          /* filename or piped-to program and arguments, subject to expansion by apr_strftime() */
    const char          *current_filename;  /* filename or piped-to program and arguments, after expansion by apr_strftime() */
    interval_t          file_interval;      /* interval of file, if rolling */

    const char          *linkname;          /* name of symbolic link to filename, subject to expansion by apr_strftime() */
    const char          *current_linkname;  /* name of symbolic link to filename, after expansion by apr_strftime() */
    interval_t          link_interval;      /* interval of symlink, if rolling */

    apr_time_t          file_expiry_time;   /* time of expiration of file */
    apr_time_t          link_expiry_time;   /* time of expiration of symlink */

    const char          *log_format_string; /* log format string */
    apr_array_header_t  *log_format_items;  /* array of function pointers */

    char                *env_condition;     /* */

    apr_file_t          *log_fd;            /* APR file descriptor */
    int                 log_type;           /* */
    server_context_t    *parent;            /* */

} log_context_t;


/* log_handler_fn_t */ 
typedef const char *log_handler_fn_t(request_rec *r, const char *s);

/* log_handler_t */ 
typedef struct log_handler_t {

    log_handler_fn_t    *function;                          /* */
    int                 use_orig_request_status_default;    /* internally redirected requests use status of original request */

} log_handler_t;


/* log_format_item -- represents '%' escape characters in a LogFormat string */
typedef struct log_format_item {

    log_handler_fn_t    *function;                  /* */
    char                *argument;                  /* */
    int                 condition_sense;            /* */
    apr_array_header_t  *conditions;                /* */
    int                 use_orig_request_status;    /* */

} log_format_item_t;


/* cached_request_time */
typedef struct cached_request_time {

    unsigned int        request_time;                                   /* */
    unsigned int        t_validate;                                     /* */
    char                request_time_string[DEFAULT_REQUEST_TIME_SIZE]; /* */

} cached_request_time_t;

static cached_request_time_t request_time_cache[TIME_CACHE_SIZE];

/* ---- end types ---- */


/* ---- begin function prototypes ---- */

/* apache hooks */
static int pre_config(apr_pool_t *p, apr_pool_t *plog, apr_pool_t *ptemp);
static void init_child(apr_pool_t *p, server_rec *s);
static int open_logs(apr_pool_t *pc, apr_pool_t *p, apr_pool_t *pt, server_rec *s);
static apr_status_t log_transaction(request_rec *r);

/* */
static apr_status_t open_log_contexts(server_rec *s, apr_pool_t *p);
static apr_status_t open_log(server_rec *s, apr_pool_t *p, log_context_t *lc);
static apr_status_t open_log_file(server_rec *s, apr_pool_t *p, log_context_t *lc);
static apr_status_t open_log_pipe(server_rec *s, apr_pool_t *p, log_context_t *lc);
static apr_status_t close_log(log_context_t *lc);
static apr_status_t update_log_symlink(server_rec *s, apr_pool_t *p, log_context_t *lc);

/* log item callbacks */
static void register_log_handler(apr_pool_t *p, const char *tag, log_handler_fn_t *log_handler_fn, int use_orig_request_status);
static const char *log_bytes_sent(request_rec *r, const char *s);
static const char *log_bytes_sent_clf(request_rec *r, const char *s);
static const char *log_connection_status(request_rec *r, const char *s);
static const char *log_constant_item(request_rec *unused, const char *s);
static const char *log_cookie(request_rec *r, const char *s);
static const char *log_env_var(request_rec *r, const char *s);
static const char *log_header_in(request_rec *r, const char *s);
static const char *log_header_out(request_rec *r, const char *s);
static char *find_multiple_header_fields(apr_pool_t *pool, const apr_table_t *table, const char *key);
static const char *log_local_address(request_rec *r, const char *s);
static const char *log_note(request_rec *r, const char *s);
static const char *log_pid_tid(request_rec *r, const char *s);
static const char *log_remote_address(request_rec *r, const char *s);
static const char *log_remote_host(request_rec *r, const char *s);
static const char *log_remote_logname(request_rec *r, const char *s);
static const char *log_remote_user(request_rec *r, const char *s);
static const char *log_request_duration(request_rec *r, const char *s);
static const char *log_request_duration_microseconds(request_rec *r, const char *s);
static const char *log_request_file(request_rec *r, const char *s);
static const char *log_request_line(request_rec *r, const char *s);
static const char *log_request_method(request_rec *r, const char *s);
static const char *log_request_protocol(request_rec *r, const char *s);
static const char *log_request_query(request_rec *r, const char *s);
static const char *log_request_time(request_rec *r, const char *s);
static const char *log_request_time_custom(request_rec *r, const char *a, apr_time_exp_t *apr_time_exp);
static const char *log_request_uri(request_rec *r, const char *s);
static const char *log_server_name(request_rec *r, const char *s);
static const char *log_server_port(request_rec *r, const char *s);
static const char *log_status(request_rec *r, const char *s);
static const char *log_virtual_host(request_rec *r, const char *s);

/* */
static char *parse_log_misc_string(apr_pool_t *p, log_format_item_t *item, const char **sa);
static const char *parse_log_format_item(apr_pool_t *p, log_format_item_t *item, const char **sa);
static const char *get_log_item_value(request_rec *r, request_rec *orig, log_format_item_t *item);
static int write_log_entry(request_rec *r, log_context_t *lc, apr_array_header_t *default_format_items);

/* directive callbacks */
static const char *directive_clf_timestamps(cmd_parms *parms, void *unused, int flag);
static const char *directive_log_format(cmd_parms *parms, void *unused, int argc, char *const argv[]);
static const char *directive_access_log(cmd_parms *parms, void *unused, int argc, char *const argv[]);
static const char *directive_transfer_log(cmd_parms *parms, void *unused, int argc, char *const argv[]);
static const char *directive_cookie_log(cmd_parms *parms, void *unused, const char *filename);

/* */
static const char *add_log_context(cmd_parms *parms, const char *filename, const char *linkname, const char *env_condition, const char *format_candidate, int log_type);

/* ---- end function prototypes ---- */


/* ---- begin global data ---- */

module AP_MODULE_DECLARE_DATA   log_access_module;  /* */

static apr_time_t               current_time;       /* */

static apr_hash_t               *log_handler_hash;  /* */
static apr_table_t              *log_format_table;  /* <format-name, format-string> pairs registered using the LogFormat directive */

static int                      mpm_is_threaded;    /* APR was compiled with thread support */
static int                      mpm_max_threads;    /* maximum thread count */

/* ---- end global data ---- */


/* ---- begin file operations ---- */


/*
 * open_log
 *
 * @param   server_rec
 * @param   apr_pool_t
 * @param   log_context_t
 *
 * @returns
 */

static apr_status_t open_log(server_rec *s, apr_pool_t *p, log_context_t *lc) { 
    
    apr_status_t    rv = APR_SUCCESS;

    if ( LOG_IS_FILE( lc->log_type ) ) {
        rv = open_log_file(s, p, lc);
        if ( lc->linkname ) {
            rv = update_log_symlink(s, p, lc);
        }
    }

    if ( LOG_IS_PIPE( lc->log_type ) ) {
        rv = open_log_pipe(s, p, lc);
    }

    if ( rv != APR_SUCCESS ) {
        ap_log_error(APLOG_MARK, APLOG_DEBUG, rv, s, "open_log(): failed to open log [%s]", lc->filename);
    }

    return rv;
}


/*
 * open_log_pipe
 *
 * @param   server_rec
 * @param   apr_pool_t
 * @param   log_context_t
 *
 */

static apr_status_t open_log_pipe(server_rec *s, apr_pool_t *p, log_context_t *lc) {

    piped_log       *pl = NULL;             /* */
    const char      *f = NULL;              /* */
    apr_status_t    rv;                     /* */
    apr_size_t      len;                    /* */
    char            filename[APR_PATH_MAX]; /* */
    apr_time_exp_t  apr_time_exp;           /* */

    /* skip pipe character and whitespace */
    for (f = lc->filename; (*f == '|') || (*f == ' ') || (*f == '\x11'); f++ ) { }

    /* */
    rv = apr_strftime(filename, &len, APR_PATH_MAX, f, &apr_time_exp);

    /* */
    if ( (pl = ap_open_piped_log(p, filename)) == NULL ) {
        ap_log_error(APLOG_MARK, APLOG_DEBUG, 0, s, "open_log_pipe(): failed to open piped log [%s] -> exec:[%s]", lc->filename, filename);
        return APR_EGENERAL;
    } else {
        rv = APR_SUCCESS;
    }

    /* save current log filename */
    lc->current_filename = apr_pstrdup(p, filename);

    /* save write descriptor */
    lc->log_fd = ap_piped_log_write_fd(pl);

    /* close read descriptor */
    rv = apr_file_close(ap_piped_log_read_fd(pl));

    return rv;
}


/*
 * open_log_file
 *
 * @param   server_rec
 * @param   apr_pool_t
 * @param   log_context_t
 *
 */

static apr_status_t open_log_file(server_rec *s, apr_pool_t *p, log_context_t *lc) {

    apr_time_t      start_of_interval;      /* */
    apr_time_t      start_of_next_interval; /* */
    char            *dirname, *sp;          /* */
    char            filename[APR_PATH_MAX]; /* */
    apr_time_exp_t  apr_time_exp;           /* */
    apr_status_t    rv = APR_SUCCESS;       /* */
    apr_size_t      len;                    /* */

    /* determine start of current interval based on current time and interval of the filename */
    current_time            = apr_time_now();
    start_of_interval       = log_timeutils_start_of_current_interval(current_time, lc->file_interval);
    start_of_next_interval  = log_timeutils_start_of_next_interval(start_of_interval, lc->file_interval); /* start_of_interval or current_time ? */

    /* format filename using start-of-interval time */
    log_timeutils_explode_localtime(&apr_time_exp, start_of_interval);
    apr_strftime(filename, &len, APR_PATH_MAX, lc->filename, &apr_time_exp);

    /* */
    if ((rv = apr_file_open(&lc->log_fd, filename, LOG_FILE_FOPEN_FLAGS, LOG_FILE_PERMISSIONS, p)) == APR_ENOENT ) {

        /* determine directory name */
        dirname = apr_pstrdup(p, filename);
        sp = ap_strrchr(dirname, '/');
        if ( sp ) {
            *sp = '\0';
        }

        /* create directory */
        if ((rv = apr_dir_make_recursive(dirname, APR_OS_DEFAULT, p)) != APR_SUCCESS) {
            ap_log_error(APLOG_MARK, APLOG_DEBUG, rv, s, "open_log_file(): could not create path [%s] for [%s]", dirname, filename);
            return rv;
        }

        /* open file */
        if ((rv = apr_file_open(&lc->log_fd, filename, LOG_FILE_FOPEN_FLAGS, LOG_FILE_PERMISSIONS, p)) != APR_SUCCESS) {
            ap_log_error(APLOG_MARK, APLOG_DEBUG, rv, s, "open_log_file(): failed to open log file [%s]", filename);
            return rv;
        }

    } 
    
    /* */
    if ( lc->log_fd ) {
        ap_log_error(APLOG_MARK, APLOG_DEBUG, rv, s, "open_log_file(): opened log [%s], APR log_fd [0x%0x]", filename, (unsigned int) lc->log_fd);
    } else {
        ap_log_error(APLOG_MARK, APLOG_DEBUG, rv, s, "open_log_file(): failed to open log [%s]", filename);
        return rv;
    }

    /* save the current log name */
    lc->current_filename    = apr_pstrdup(p, filename);
    lc->file_expiry_time    = start_of_next_interval-1;

    /* */
    return APR_SUCCESS;
}


/*
 * update_log_symlink
 *
 * @param   server_rec
 * @param   apr_pool_t
 * @param   log_context_t
 *
 */

static apr_status_t update_log_symlink(server_rec *s, apr_pool_t *p, log_context_t *lc) {

    apr_status_t    rv;                     /* */
    char            linkname[APR_PATH_MAX]; /* */
    apr_size_t      len;                    /* */
    apr_finfo_t     finfo;                  /* */
    apr_time_exp_t  apr_time_exp;           /* */
    apr_time_t      start_of_interval;      /* */
    apr_time_t      start_of_next_interval; /* */
    char            *dirname, *sp;          /* */

    /* */
    if ( ! LOG_IS_SYMLINKED(lc->log_type) ) {
        return APR_SUCCESS;
    }

    /* remove existing link */
    if ( (rv = apr_stat(&finfo, lc->current_linkname, APR_FINFO_LINK, p)) == APR_SUCCESS ) {
        if ( finfo.filetype == APR_LNK ) {
            if ( (rv = unlink( lc->current_linkname )) == 0 ) {
                ap_log_error(APLOG_MARK, APLOG_DEBUG, rv, s, "update_log_symlink(): removed symbolic link [%s]", lc->current_linkname);
            } else {
                ap_log_error(APLOG_MARK, APLOG_DEBUG, rv, s, "update_log_symlink(): failed to remove symbolic link [%s]", lc->current_linkname);
            }
        } else {
            ap_log_error(APLOG_MARK, APLOG_DEBUG, rv, s, "update_log_symlink(): [%s] is not a symbolic link; finfo.filetype = [%d]", lc->current_linkname, finfo.filetype);
            return APR_EGENERAL;
        }
    }

    /* determine start of current interval based on current time and interval of the symlink */
    start_of_interval       = log_timeutils_start_of_current_interval(current_time, lc->link_interval);
    start_of_next_interval  = log_timeutils_start_of_next_interval(start_of_interval, lc->link_interval); /* start_of_interval or current_time ? */

    /* format name of new symlink using start-of-interval time */
    log_timeutils_explode_localtime(&apr_time_exp, start_of_interval);
    apr_strftime(linkname, &len, APR_PATH_MAX, lc->linkname, &apr_time_exp);

    /* determine path component of new symlink */
    dirname = apr_pstrdup(p, linkname);
    sp = ap_strrchr(dirname, '/');
    if ( sp ) {
        *sp = '\0';
    }

    /* create directory for symlink */
    if ((rv = apr_dir_make_recursive(dirname, APR_OS_DEFAULT, p)) != APR_SUCCESS) {
        ap_log_error(APLOG_MARK, APLOG_DEBUG, rv, s, "update_log_symlink(): could not create directory [%s] for symbolic link [%s]", dirname, linkname);
        return rv;
    }

    /* create new symlink */
    if ((rv = symlink( lc->current_filename, linkname)) == 0) {
        ap_log_error(APLOG_MARK, APLOG_DEBUG, rv, s, "update_log_symlink(): created symbolic link [%s] -> [%s]", linkname, lc->current_filename);
    } else {
        if ( errno != EEXIST ) {
            ap_log_error(APLOG_MARK, APLOG_DEBUG, rv, s, "update_log_symlink(): could not create symbolic link [%s] -> [%s]: %s", linkname, lc->current_filename, strerror(errno));
            return APR_EGENERAL;
        }
    }

    /* save the current symlink name */
    lc->current_linkname = apr_pstrdup(p, linkname);
    lc->link_expiry_time = start_of_next_interval-1;

    /* */
    return APR_SUCCESS;
}


/*
 * close_log
 *
 * @param   log_context_t
 * @return  apr_status_t
 *
 */

static apr_status_t close_log(log_context_t *lc) {

    apr_status_t    rv = APR_SUCCESS;

    if ( lc->log_fd == NULL ) {
        return APR_SUCCESS;
    }

    if ((rv = apr_file_close(lc->log_fd)) != APR_SUCCESS ) {
        ap_log_perror(APLOG_MARK, APLOG_DEBUG, rv, NULL, "close_log(): problem closing log [%s], APR log_fd [0x%0x]", lc->current_filename, (unsigned int) lc->log_fd);
        return rv;
    } else {
        ap_log_perror(APLOG_MARK, APLOG_DEBUG, rv, NULL, "close_log(): [%s], APR log_fd [0x%0x]", lc->current_filename, (unsigned int) lc->log_fd);
    }

    return APR_SUCCESS;
}


/* ---- end file operations ---- */




/* ---- begin log item callbacks ---- */

/*
 * @param   request_rec
 * @return  const char *
 *
 */

static const char *log_remote_host(request_rec *r, const char *s) {
    return ap_escape_logitem(r->pool, ap_get_remote_host(r->connection, r->per_dir_config, REMOTE_NAME, NULL));
}

static const char *log_remote_address(request_rec *r, const char *s) {
    return r->connection->remote_ip;
}

static const char *log_local_address(request_rec *r, const char *s) {
    return r->connection->local_ip;
}

static const char *log_remote_logname(request_rec *r, const char *s) {
    return ap_escape_logitem(r->pool, ap_get_remote_logname(r));
}

static const char *log_remote_user(request_rec *r, const char *s) {
    const char *rvalue = r->user;

    if (rvalue == NULL) {
        rvalue = "-";
    } else if (strlen(rvalue) == 0) { /* no apr_strlen() */
        rvalue = "\"\"";
    } else {
        rvalue = ap_escape_logitem(r->pool, rvalue);
    }

    return rvalue;
}

static const char *log_request_line(request_rec *r, const char *s) {

    /*
    if the original request contained a password, re-write to contain xxxxxx instead:
    the truncation before the protocol string for HTTP/0.9 requests.
    r->the_request contains the unmodified request.
    */

    return ap_escape_logitem(r->pool, (r->parsed_uri.password) ? apr_pstrcat(r->pool, r->method, " ", apr_uri_unparse(r->pool, &r->parsed_uri, 5), r->assbackwards ? NULL : " ", r->protocol, NULL) : r->the_request);
}

static const char *log_request_file(request_rec *r, const char *s) {
    return ap_escape_logitem(r->pool, r->filename);
}

static const char *log_request_uri(request_rec *r, const char *s) {
    return ap_escape_logitem(r->pool, r->uri);
} 

static const char *log_request_method(request_rec *r, const char *s) {
    return ap_escape_logitem(r->pool, r->method);
}

static const char *log_request_protocol(request_rec *r, const char *s) {
    return ap_escape_logitem(r->pool, r->protocol);
}

static const char *log_request_query(request_rec *r, const char *s) {
    return (r->args) ? apr_pstrcat(r->pool, "?", ap_escape_logitem(r->pool, r->args), NULL) : "";
}

static const char *log_status(request_rec *r, const char *s) {
    if ( r->status <= 0 ) {
        return "-";
    } else {
        return apr_itoa(r->pool, r->status);
    }
}

static const char *log_bytes_sent_clf(request_rec *r, const char *s) {
    if ( (r->sent_bodyct == 0) || (r->bytes_sent == 0) ) {
        return "-";
    } else {
        return apr_off_t_toa(r->pool, r->bytes_sent);
    }
}

static const char *log_bytes_sent(request_rec *r, const char *s) {
    if ( (r->sent_bodyct == 0) || (r->bytes_sent == 0) ) {
        return "0";
    } else {
        return apr_off_t_toa(r->pool, r->bytes_sent);
    }
}

static const char *log_header_in(request_rec *r, const char *s) {
    return ap_escape_logitem(r->pool, apr_table_get(r->headers_in, s));
}

static const char *log_header_out(request_rec *r, const char *s) {
    const char *cp = NULL;

    if (!strcasecmp(s, "content-type") && r->content_type) {
        cp = ap_field_noparam(r->pool, r->content_type);
    } else if (!strcasecmp(s, "set-cookie")) {
        cp = find_multiple_header_fields(r->pool, r->headers_out, s);
    } else {
        cp = apr_table_get(r->headers_out, s);
    }

    return ap_escape_logitem(r->pool, cp);
}

/*
 * find_multiple_header_fields
 *
 * @param   apr_pool_t
 * @param   apr_table_t
 * @param   const char*
 */

static char *find_multiple_header_fields(apr_pool_t *pool, const apr_table_t *table, const char *key) {

    const apr_array_header_t    *header_fields;
    const apr_table_entry_t     *t_element;
    const apr_table_entry_t     *t_end;
    apr_size_t                  len;

    typedef struct sle {
        struct sle  *next;
        const char  *value;
        apr_size_t  len;
    } sle_t;
    sle_t *result_list, *rp;

    header_fields = apr_table_elts(table);

    if (!header_fields->nelts) {
        return NULL;
    }

    t_element = (const apr_table_entry_t *) header_fields->elts;
    t_end = t_element + header_fields->nelts;

    result_list = NULL;
    rp = NULL;
    len = 1;    /* '\0' */

    do {
        if (!strcasecmp(t_element->key, key)) {
            if (!result_list) {
                result_list = rp = apr_palloc(pool, sizeof(struct sle));
            } else {
                rp = rp->next = apr_palloc(pool, sizeof(struct sle));
                len += 2;   /* ", " */
            }

            rp->next  = NULL;
            rp->value = t_element->val;
            rp->len   = strlen(rp->value); /* no apr_strlen() */

            len += rp->len;
        }
        ++t_element;
    } while (t_element < t_end);

    if ( result_list ) {
        char *result = apr_palloc(pool, len);
        char *cp = result;

        rp = result_list;
        while (rp) {
            if (rp != result_list) {
                *cp++ = ',';
                *cp++ = ' ';
            }
            memcpy(cp, rp->value, rp->len);
            cp += rp->len;
            rp = rp->next;
        }
        *cp = '\0';

        return result;
    }

    return NULL;
}

static const char *log_note(request_rec *r, const char *s) {
    return ap_escape_logitem(r->pool, apr_table_get(r->notes, s));
}

static const char *log_env_var(request_rec *r, const char *s) {
    return ap_escape_logitem(r->pool, apr_table_get(r->subprocess_env, s));
}

static const char *log_cookie(request_rec *r, const char *s) {
    const char *cookies         = NULL;
    const char *cookie_start    = NULL;

    /* */
    if ( (cookies = apr_table_get(r->headers_in, "Cookie")) ) {
        if ( (cookie_start = ap_strstr_c(cookies, s)) ) {
            char *cookie, *cookie_end;
            cookie_start += strlen(s) + 1;  /* cookie_name + '=' */ /* no apr_strlen() */
            cookie     = apr_pstrdup(r->pool, cookie_start);
            cookie_end = ap_strchr(cookie, ';');
            if (cookie_end) {
                *cookie_end = '\0';
            }
            return ap_escape_logitem(r->pool, cookie);
        }
    }

    /* */
    return NULL;
}

static const char *log_request_time_custom(request_rec *r, const char *s, apr_time_exp_t *apr_time_exp) {

    char        timestring[MAX_STRING_LEN];
    apr_size_t  retval;

    apr_strftime(timestring, &retval, MAX_STRING_LEN, s, apr_time_exp);

    return apr_pstrdup(r->pool, timestring);
}

static const char *log_request_time(request_rec *r, const char *s) {

    server_context_t    *sc = NULL;
    apr_time_exp_t      apr_time_exp;
    
    sc = ap_get_module_config(r->server->module_config, &log_access_module);

    if (s && *s) {  /* custom format */

        /* if CLF timestamps are required, use the current time */
        log_timeutils_explode_localtime(&apr_time_exp, sc->clf_timestamps ? current_time : r->request_time);

        return log_request_time_custom(r, s, &apr_time_exp);
    }

    /*
    XXX -- not sure of the value of this -- gjvc
    this uses the same technique as ap_explode_recent_localtime()
    optimistic caching with logic to detect and correct race conditions.
    see the comments in server/util_time.c for more information.
    */

    /* */
    cached_request_time_t   *cached_time = apr_palloc(r->pool, sizeof(cached_request_time_t));
    apr_time_t              request_time = sc->clf_timestamps ? current_time : r->request_time;
    unsigned int            t_seconds    = (unsigned int) apr_time_sec(request_time);
    unsigned int            cache_slot   = t_seconds & TIME_CACHE_MASK;

    /* */
    *cached_time = request_time_cache[cache_slot];

    /* if snapshot is invalid or old, compute proper time string and cache item */
    if ((t_seconds != cached_time->request_time) || (t_seconds != cached_time->t_validate)) {

        char    sign;
        int     timz;

        ap_explode_recent_localtime(&apr_time_exp, request_time);
        timz = apr_time_exp.tm_gmtoff;
        if (timz < 0) {
            timz = -timz;
            sign = '-';
        } else {
            sign = '+';
        }

        /* cache time string and value */
        apr_snprintf( cached_time->request_time_string,
                      DEFAULT_REQUEST_TIME_SIZE,
                      "[%02d/%s/%d:%02d:%02d:%02d %c%.2d%.2d]",
                      apr_time_exp.tm_mday,
                      apr_month_snames[apr_time_exp.tm_mon],
                      apr_time_exp.tm_year + 1900,
                      apr_time_exp.tm_hour,
                      apr_time_exp.tm_min,
                      apr_time_exp.tm_sec,
                      sign,
                      (timz / (60 * 60)), 
                      (timz % (60 * 60)) / 60
                    );

        cached_time->request_time   = t_seconds;
        cached_time->t_validate     = t_seconds;

        request_time_cache[cache_slot]  = *cached_time;
    }

    return cached_time->request_time_string;
}

static const char *log_request_duration(request_rec *r, const char *s) {
    return apr_psprintf(r->pool, "%" APR_TIME_T_FMT, apr_time_sec( current_time - r->request_time ));
}

static const char *log_request_duration_microseconds(request_rec *r, const char *s) {
    return apr_psprintf(r->pool, "%" APR_TIME_T_FMT, current_time - r->request_time);
}

static const char *log_virtual_host(request_rec *r, const char *s) {
    return ap_escape_logitem(r->pool, r->server->server_hostname);
}

static const char *log_server_port(request_rec *r, const char *s) {
    return apr_psprintf(r->pool, "%u", r->server->port ? r->server->port : ap_default_port(r));
}

static const char *log_server_name(request_rec *r, const char *s) {
    return ap_escape_logitem(r->pool, ap_get_server_name(r));
}

static const char *log_pid_tid(request_rec *r, const char *s) {

    if (*s == '\0' || !strcmp(s, "pid")) {
        return ap_append_pid(r->pool, "", "");

    } else if (!strcmp(s, "tid") || !strcmp(s, "hextid")) {

        apr_os_thread_t tid = 0;

        if ( mpm_is_threaded ) {
            tid = apr_os_thread_current();
        }

#if ( APR_MAJOR_VERSION > 1 ) || (APR_MAJOR_VERSION == 1 && APR_MINOR_VERSION >= 2)
        /* APR >= 1.2 can format a thread id in hex */
        return apr_psprintf(r->pool, *s == 'h' ? "%pt" : "%pT", (void *) &tid);
#else
        /* APR < 1.2 cannot */
        return apr_psprintf(r->pool, "%pT", (void *) &tid);
#endif

    }

    /* bogus format */
    return s;
}

static const char *log_connection_status(request_rec *r, const char *s) {

    if ( r->connection->aborted )
        return "X";

    if ( (r->connection->keepalive == AP_CONN_KEEPALIVE) && 
         (!r->server->keep_alive_max || (r->server->keep_alive_max - r->connection->keepalives) > 0)) {
        return "+";
    }

    return "-";
}

static const char *log_constant_item(request_rec * unused, const char *s) {
    return s;
}

/* ---- end log item callbacks ---- */


/*
 * format_items_from_format_string
 *
 * @param   apr_pool_t
 * @param   const char *
 * @param   const char **
 *
 * @returns apr_array_header_t of log_format_item_t
 */

static apr_array_header_t *format_items_from_format_string(apr_pool_t *p, const char *s, const char **err) {

    apr_array_header_t  *log_format_items = apr_array_make(p, 32, sizeof(log_format_item_t));
    const char          *result = NULL;

    /* */
    while ( s && *s ) {
        if ((result = parse_log_format_item(p, (log_format_item_t *) apr_array_push(log_format_items), &s)) != NULL) {
            *err = result;
            return NULL;
        }
    }

    /* */
    s = APR_EOL_STR;
    parse_log_format_item(p, (log_format_item_t *) apr_array_push(log_format_items), &s);

    /* */
    return log_format_items;
}


/*
 * parse_log_format_item
 *
 * @param   apr_pool_t
 * @param   log_format_item_t
 * @param   const char **
 *
 */

static const char *parse_log_format_item(apr_pool_t * p, log_format_item_t * item, const char **sa) {

    log_handler_t   *handler = NULL;
    const char      *s       = *sa;
    const char      *s_orig  = *sa;

    if ( *s++ != '%' ) {
        return parse_log_misc_string(p, item, sa);
    }

    item->condition_sense   = 0;
    item->conditions        = NULL;

    if ( *s == '%' ) {
        item->function = log_constant_item;
        item->argument = apr_pstrdup(p, "%");
        *sa = ++s;

        return NULL;
    }

    item->argument = NULL; /* NULL or empty string? */
    item->use_orig_request_status = -1;

    while (*s) {
        int i;

        switch (*s) {
            case '!':
                ++s;
                item->condition_sense = !item->condition_sense;
                break;

            case '<':
                ++s;
                item->use_orig_request_status = 1;
                break;

            case '>':
                ++s;
                item->use_orig_request_status = 0;
                break;

            case ',':
                ++s;
                break;

            case '{':
                ++s;
                item->argument = ap_getword(p, &s, '}');
                break;

            case '0':
            case '1':
            case '2':
            case '3':
            case '4':
            case '5':
            case '6':
            case '7':
            case '8':
            case '9':
                i = *s - '0';
                while (apr_isdigit(*++s)) {
                    i = i * 10 + (*s) - '0';
                }
                if ( item->conditions == NULL ) {
                    item->conditions = apr_array_make(p, 4, sizeof(int));
                }
                *(int *)apr_array_push(item->conditions) = i;
                break;

            default:
                handler = (log_handler_t *) apr_hash_get(log_handler_hash, s++, 1);
                if ( handler == NULL ) {
                    char unused[2];

                    unused[0] = s[-1];
                    unused[1] = '\0';
                    return apr_pstrcat(p, "Unknown directive [%", unused, "] passed to LogFormat", NULL);
                }
                item->function = handler->function;
                if (item->use_orig_request_status == -1) {
                    item->use_orig_request_status = handler->use_orig_request_status_default;
                }
                *sa = s;
                return NULL;
        }
    }

    /* */
    ap_log_perror(APLOG_MARK, APLOG_CRIT, 0, NULL, "parse_log_format_item: runaway LogFormat escape sequence [%s]", s_orig);

    /* */
    return NULL;
}


/*
 * parse_log_misc_string -- parse a misc log format string
 *
 * @param   apr_pool_t
 * @parama  log_format_item_t
 * @param   const char **
 * @return  char *
 */

static char *parse_log_misc_string(apr_pool_t *p, log_format_item_t *item, const char **sa) {

    const char  *s; /* pointer to source */
    char        *d; /* pointer to destination */

    s = *sa;
    while (*s && *s != '%') {
        s++;
    }

    /* might allocate a few bytes extra if there is a backslash escape in the format string */
    item->function      = log_constant_item;
    item->argument      = apr_palloc(p, s - *sa + 1);
    item->conditions    = NULL;

    s = *sa;
    d = item->argument;
    while (*s && *s != '%') {

        if (*s == '\\') {
            s++;
            switch (*s) {
                case '\\':
                    *d++ = '\\';
                    s++;
                    break;

                case 'r':
                    *d++ = '\r';
                    s++;
                    break;

                case 'n':
                    *d++ = '\n';
                    s++;
                    break;

                case 't':
                    *d++ = '\t';
                    s++;
                    break;

                default:
                    /* copy verbatim */
                    *d++ = '\\';
                    /* allow the loop to deal with this *s in the normal fashion so that it handles end of string etc. properly */
                    break;
            }
        } else {
            *d++ = *s++;
        }
    }

    *sa = s;
    *d = '\0';

    return NULL;
}


/*
 * init_server_config
 *
 * @param   apr_pool_t
 * @param   server_rec
 *
 * @returns void*
 */

static void *init_server_config(apr_pool_t *p, server_rec *s) {

    server_context_t    *child  = NULL;

    ap_log_error(APLOG_MARK, APLOG_DEBUG, 0, s, "init_server_config() [%s]", s->server_hostname);

    /* allocate */
    child = (server_context_t *) apr_palloc(p, sizeof(server_context_t));

    /* initialise */
    child->server_hostname          = apr_pstrdup(p, s->server_hostname);
    child->clf_timestamps           = 1;
    child->default_format_items     = NULL;
    child->default_format_string    = NULL;
    child->log_contexts             = apr_array_make(p, 1, sizeof(log_context_t));
    child->error_log_contexts       = apr_array_make(p, 1, sizeof(log_context_t));
    child->log_format_table         = apr_table_make(p, 4);
    child->parent_log_contexts      = NULL;

    /* */
    return child;
}


/*
 * merge_server_config
 *
 * @param   apr_pool_t
 * @param   void *
 * @param   void *
 *
 * @returns void *
 * @comment     add a pointer from the vhost log state to the list of logs specified for the non-vhost configuration.
 *              vhosts inherit any globally-defined format names.
 *              this does the right thing, but child->log_formats is getting overwritten somewhere
 *
 */

static void *merge_server_config(apr_pool_t *p, void *parent_v, void *child_v) {

    server_context_t *parent = (server_context_t *) parent_v;
    server_context_t *child  = (server_context_t *) child_v;

    ap_log_perror(APLOG_MARK, APLOG_DEBUG, 0, p, "merge_server_config(): [%s] -> [%s]", parent->server_hostname, child->server_hostname);

    child->parent_log_contexts = parent->log_contexts;

    if ( child->default_format_items == NULL ) {
        child->default_format_string = parent->default_format_string;
        child->default_format_items  = parent->default_format_items;
    }

    child->log_format_table = apr_table_overlay(p, parent->log_format_table, child->log_format_table);

    return child;
}


/*
 * open_log_contexts
 *
 * @param   server_rec
 * @param   apr_pool_t
 *
 * @returns apr_status_t
 */

static apr_status_t open_log_contexts(server_rec *s, apr_pool_t *p) {

    int                 i               = 0;
    apr_status_t        rv              = APR_SUCCESS;
    log_context_t       *lc             = NULL;
    log_context_t       *lc_array       = NULL;
    server_context_t    *sc             = NULL;
    apr_array_header_t  *log_contexts   = NULL;
    const char          *error          = NULL;

    /* */
    sc = ap_get_module_config(s->module_config, &log_access_module);

    /* */
    if ( sc->default_format_string ) {
        sc->default_format_items = format_items_from_format_string(p, sc->default_format_string, &error);
    }

    /* */
    if ( sc->default_format_items == NULL ) {
        sc->default_format_string = apr_pstrdup(p, LOG_FORMAT_CLF);
        sc->default_format_items  = format_items_from_format_string(p, LOG_FORMAT_CLF, &error);
    }

    /* */
    if ( sc->log_contexts ) {
        log_contexts = sc->log_contexts;
    } else if ( sc->parent_log_contexts ) {
        log_contexts = sc->parent_log_contexts;
    } else {
        log_contexts = NULL;
    }

    /* */
    if ( log_contexts ) {

        lc_array = (log_context_t *) log_contexts->elts;

        for (i = 0; i < log_contexts->nelts; ++i) {
            lc = &lc_array[i];
            if ( (rv = open_log(s, p, lc)) != APR_SUCCESS ) {
                return rv;
            }
        }
    }

    /* */
    if ( sc->error_log_contexts ) {
        log_contexts = sc->error_log_contexts;
    } else {
        log_contexts = NULL;
    }

    /* */
    if ( log_contexts ) {

        lc_array = (log_context_t *) log_contexts->elts;

        for (i = 0; i < log_contexts->nelts; ++i) {
            lc = &lc_array[i];
            if ( (rv = open_log(s, p, lc)) != APR_SUCCESS ) {
                return rv;
            }
        }
    }

    /* */
    return APR_SUCCESS;
}

/* ---- end ---- */


/* ---- begin hooks ---- */


/*
 * pre_config
 *
 * @param   apr_pool_t*
 * @param   apr_pool_t*
 * @param   apr_pool_t*
 *
 * @return  int
 */

static int pre_config(apr_pool_t *p, apr_pool_t *plog, apr_pool_t *ptemp) {

    /* init log handler hash */
    log_handler_hash = apr_hash_make(p);

    /* register handlers */
    register_log_handler(p, "a", log_remote_address, 0);
    register_log_handler(p, "A", log_local_address, 0);
    register_log_handler(p, "b", log_bytes_sent_clf, 0);
    register_log_handler(p, "B", log_bytes_sent, 0);
    register_log_handler(p, "C", log_cookie, 0);
    register_log_handler(p, "D", log_request_duration_microseconds, 1);
    register_log_handler(p, "e", log_env_var, 0);
    register_log_handler(p, "f", log_request_file, 0);
    register_log_handler(p, "h", log_remote_host, 0);
    register_log_handler(p, "H", log_request_protocol, 0);
    register_log_handler(p, "i", log_header_in, 0);
    register_log_handler(p, "l", log_remote_logname, 0);
    register_log_handler(p, "m", log_request_method, 0);
    register_log_handler(p, "n", log_note, 0);
    register_log_handler(p, "o", log_header_out, 0);
    register_log_handler(p, "p", log_server_port, 0);
    register_log_handler(p, "P", log_pid_tid, 0);
    register_log_handler(p, "q", log_request_query, 0);
    register_log_handler(p, "r", log_request_line, 1);
    register_log_handler(p, "s", log_status, 1);
    register_log_handler(p, "t", log_request_time, 0);
    register_log_handler(p, "T", log_request_duration, 1);
    register_log_handler(p, "u", log_remote_user, 0);
    register_log_handler(p, "U", log_request_uri, 1);
    register_log_handler(p, "v", log_virtual_host, 0);
    register_log_handler(p, "V", log_server_name, 0);
    register_log_handler(p, "X", log_connection_status, 0);

    /* and that's that */
    return APR_SUCCESS;
}


/*
 * register_log_handler
 *
 * @param
 * @param
 * @param
 *
 */

static void register_log_handler(apr_pool_t * p, const char *tag, log_handler_fn_t * log_handler_fn, int use_orig_request_status) {

    log_handler_t *log_handler = NULL;

    /* allocate */
    log_handler = apr_palloc(p, sizeof(log_handler_t));

    /* initialise */
    log_handler->function                           = log_handler_fn;           /* */
    log_handler->use_orig_request_status_default    = use_orig_request_status;  /* internally redirected requests return status of original request */

    /* insert */
    apr_hash_set(log_handler_hash, tag, 1, (const void *)log_handler);
}


/*
 * init_child
 *
 * @param   apr_pool_t*
 * @param   server_rec*
 *
 * @return  void
 */

static void init_child(apr_pool_t* p, server_rec* s) {
    apr_status_t rv;

    rv = ap_mpm_query(AP_MPMQ_IS_THREADED, &mpm_is_threaded);
    if ( (rv == APR_SUCCESS) && (mpm_is_threaded == 1) ) {
        rv = ap_mpm_query(AP_MPMQ_MAX_THREADS, &mpm_max_threads);
    }

}


/*
 * open_logs
 *
 * @param   apr_pool_t*
 * @param   apr_pool_t*
 * @param   apr_pool_t*
 * @param   server_rec*
 *
 * @return  int
 */

static int open_logs(apr_pool_t *pconf, apr_pool_t *plog, apr_pool_t *ptemp, server_rec *s) {

    apr_status_t    rv = APR_SUCCESS;
    server_rec      *server = s;

    /* */
    current_time = apr_time_now();


    /* */
    for (server = s; (rv == APR_SUCCESS) && server; server = server->next) {
        if ( ( rv = open_log_contexts(server, server->process->pool) ) != APR_SUCCESS ) {
            return rv;
        }
    }

    /* */
    return rv;
}


/*
 * write_log_entry
 *
 * @param   request_rec*
 * @param   log_context_t*
 * @param   apr_array_header_t*
 *
 * @return  int
 */

static int write_log_entry(request_rec *r, log_context_t *lc, apr_array_header_t *default_format_items) {

    apr_status_t        rv = APR_SUCCESS;
    apr_array_header_t  *log_format_items;
    log_format_item_t   *items;
    apr_size_t          len = 0;
    request_rec         *orig;
    char                *env_condition;
    const char          **strs;
    int                 *strl;
    int                 i;
    char                *log_entry;
    char                *s;

    /* */
    if ( lc->filename == NULL ) {
        return DECLINED;
    }

    /* check conditional environment variable-controlled logging decisions to make */
    if ( lc->env_condition ) {
        env_condition = lc->env_condition;
        if ( *env_condition != '!' ) {
            if ( apr_table_get(r->subprocess_env, env_condition) == NULL) {
                return DECLINED;
            }
        } else {
            if ( apr_table_get(r->subprocess_env, &env_condition[1]) != NULL ) {
                return DECLINED;
            }
        }
    }

    /* request not declined */
    current_time = apr_time_now();

    /* */
    log_format_items = lc->log_format_items ? lc->log_format_items : default_format_items;

    /* */
    strs = apr_palloc(r->pool, sizeof(char *) * (log_format_items->nelts));
    strl = apr_palloc(r->pool, sizeof(int)    * (log_format_items->nelts));
    items = (log_format_item_t *) log_format_items->elts;

    /* find original request */
    orig = r;
    while ( orig->prev ) {
        orig = orig->prev;
    }

    /* find final request */
    while ( r->next ) {
        r = r->next;
    }

    for ( i = 0; i < log_format_items->nelts; ++i ) {
        strs[i] = get_log_item_value(r, orig, &items[i]);
    }

    for ( i = 0; i < log_format_items->nelts; ++i ) {
        len += strl[i] = strlen( strs[i] ); /* no apr_strlen() */
    }

    /* */
    if ( LOG_IS_ROLLING( lc->log_type ) ) {

        /* open new log file if necessary */
        if ( current_time >= lc->file_expiry_time ) {
            close_log(lc);
            open_log(r->server, r->server->process->pool, lc);
        }

        /* update symlink if necessary */
        if ( current_time >= lc->link_expiry_time ) {
            update_log_symlink(r->server, r->server->process->pool, lc);
        }

    }

    /* construct log entry string */
    log_entry = apr_palloc(r->pool, len + 1);
    for (i = 0, s = log_entry; i < log_format_items->nelts; ++i) {
        memcpy(s, strs[i], strl[i]);
        s += strl[i];
    }

    /* write to file */
    if (( rv = apr_file_write(lc->log_fd, log_entry, &len)) == APR_SUCCESS ) {
        rv = APR_SUCCESS;
    } else {
        rv = DECLINED;
    }

    /* */
    return rv;
}


/*
 * get_log_item_value
 *
 * @param       request_rec
 * @param       request_rec
 * @param       log_format_item_t
 *
 * @returns     const char *
 */

static const char *get_log_item_value(request_rec *r, request_rec * orig, log_format_item_t * item) {

    const char *cp;

    if ( item->conditions && item->conditions->nelts != 0 ) {
        int *conds = (int *)item->conditions->elts;
        int in_list = 0;
        int i;

        for (i = 0; i < item->conditions->nelts; ++i) {
            if (r->status == conds[i]) {
                in_list = 1;
                break;
            }
        }

        if ( (item->condition_sense && in_list) || (!item->condition_sense && !in_list) ) {
            return "-";
        }
    }

    cp = (const char *) ((*item->function) (item->use_orig_request_status ? orig : r, item->argument));

    return cp ? cp : "-";
}


/*
 * log_transaction
 *
 * @param       request_rec
 *
 * @returns     apr_status_t
 *
 */
static apr_status_t log_transaction(request_rec *r) {

    server_context_t    *sc;
    log_context_t       *lc_array;
    apr_array_header_t  *log_contexts;
    int                 i;

    /* */
    sc = ap_get_module_config(r->server->module_config, &log_access_module);

    /* */
    if ( sc->log_contexts ) {
        log_contexts = sc->log_contexts;
    } else if ( sc->parent_log_contexts ) {
        log_contexts = sc->parent_log_contexts;
    } else {
        log_contexts = NULL;
    }

    /* */
    if ( log_contexts ) {
        lc_array = (log_context_t *) log_contexts->elts;
        for (i = 0; i < log_contexts->nelts; ++i) {
            write_log_entry(r, &lc_array[i], sc->default_format_items);
        }
    }

    /* */
    return APR_SUCCESS;
}

/* ---- end hooks ---- */


/*
 * register_hooks
 *
 * @param       apr_pool_t
 * @return      void
 *
 */

static void register_hooks(apr_pool_t *p) {

    /* init hooks */
    ap_hook_pre_config(pre_config, NULL, NULL, APR_HOOK_REALLY_FIRST);
    ap_hook_child_init(init_child, NULL, NULL, APR_HOOK_MIDDLE);
    ap_hook_open_logs(open_logs, NULL, NULL, APR_HOOK_MIDDLE);
    ap_hook_log_transaction(log_transaction, NULL, NULL, APR_HOOK_MIDDLE);

    /* log_format_table */
    log_format_table = apr_table_make(p, 4);
    apr_table_set(log_format_table, "CLF", LOG_FORMAT_CLF );
}


/* ---- begin directive processing callbacks ---- */

/*
 * directive_log_format
 *
 * @directive   LogFormat
 * @parameters  <FORMAT_STRING> [<FORMAT_NAME>]
 * @description set default logfile format, or define a custom named format string
 *
 * @param       cmd_parms
 * @param       void*
 * @param       const char*
 * @param       const char*[]
 * @returns     const char*
 */
static const char *directive_log_format(cmd_parms *parms, void *unused, int argc, char *const argv[]) {

    server_context_t    *sc = NULL;
    const char          *error = NULL;

    sc = ap_get_module_config(parms->server->module_config, &log_access_module);

    switch ( argc ) {
        case 1:
            sc->default_format_string = apr_pstrdup(parms->pool, argv[0]);
            sc->default_format_items  = format_items_from_format_string(parms->pool, sc->default_format_string, &error);
            break;

        case 2:
            apr_table_set( log_format_table, argv[1], argv[0] );
            apr_table_set( sc->log_format_table, argv[1], argv[0] );
            break;

        default:
            error = "syntax: LogFormat <FORMAT_STRING> [<FORMAT_NAME>]";
            break;
    }

    return error;
}


/*
 * directive_access_log
 *
 * @directive       RollingLog
 * @parameters      <FILENAME> <LINKNAME> [<FORMAT>|<FORMAT_NAME> [<CONDITION>]]
 * @description     add a log to a virtual host / server configuration
 *
 * @param           cmd_parms*
 * @param           void*
 * @param           int
 * @param           char * const argv[]
 * @returns         const char*
 */

static const char *directive_access_log(cmd_parms *parms, void *unused, int argc, char *const argv[]) {
 
    int             i;
    int             log_type = 0;
    const char      *env_condition = NULL;
    const char      *format_candidate = NULL;
    const char      *linkname = NULL;
    const char      *logname = NULL;
    char            filename[APR_PATH_MAX];
    apr_size_t      len;
    apr_status_t    rv;
    const char      *tmp = NULL;
    apr_time_exp_t  apr_time_exp;

    /* */
    if ( argc < 2 || argc > 4 ) {
        return "syntax: AccessLog <FILENAME> [symlink=<LINKNAME>] [format=<FORMAT>|<FORMAT_NAME>] [env=<CONDITION>]";
    }

    /* dump command line */
    for ( i=0; i < argc; i++ ) {
        ap_log_perror(APLOG_MARK, APLOG_DEBUG, 0, parms->pool, "directive_access_log argv[%d/%d] = [%s]", i, argc, argv[i]);
    }

    /* */
    for ( i=0; i < argc; i++ ) {

        if ( linkname == NULL ) {
            tmp = argv[i];
            if ( strncasecmp(tmp, "symlink=", 8) == 0) {
                linkname = &tmp[8];

                rv = apr_strftime(filename, &len, APR_PATH_MAX, linkname, &apr_time_exp);
                if ( strcmp(linkname, filename) ) { /* unequal strings, so linkname is a strftime() format string */
                    log_type |= LOG_TYPE_ROLLING;
                }
                continue;
            }
        }

        if ( format_candidate == NULL ) {
            tmp = argv[i];
            if ( strncasecmp(tmp, "format=", 7) == 0) {
                format_candidate = &tmp[7];
                continue;
            }
        }

        if ( env_condition == NULL ) {
            tmp = argv[i];
            if ( strncasecmp(tmp, "env=", 4) == 0) {
                if ((tmp[4] == '\0') || ((tmp[4] == '!') && (tmp[5] == '\0'))) {
                    return "missing environment variable name";
                } else {
                    env_condition = &tmp[4];
                    log_type |= LOG_TYPE_CONDITIONAL;
                    continue;
                }
            }
        }

        if ( logname == NULL ) {
            logname = argv[i];
            if ( logname[0] == '|' ) {
                log_type |= LOG_TYPE_PIPE;
            } else {
                log_type |= LOG_TYPE_FILE;
            }
            rv = apr_strftime(filename, &len, APR_PATH_MAX, logname, &apr_time_exp);
            if ( strcmp(logname, filename) ) { /* unequal strings, so linkname is a strftime() format string */
                log_type |= LOG_TYPE_ROLLING;
            }
        }
    }

    /* */
    return add_log_context(parms, logname, linkname, env_condition, format_candidate, log_type);
}


/*
 * add_log_context
 *
 * @description     add a log to a virtual host / server configuration
 *
 * @param           cmd_parms*
 * @param           char*
 * @param           char*
 * @param           char*
 * @param           char*
 * @param           int
 * @returns         const char*
 */

static const char *add_log_context(cmd_parms *parms, const char *filename, const char *linkname, const char *env_condition, const char *format_candidate, int log_type) {

    server_context_t    *sc = NULL;
    log_context_t       *lc = NULL;
    const char          *error          = NULL;
    const char          *log_format_string  = NULL;

    /* */
    ap_log_perror(APLOG_MARK, APLOG_DEBUG, 0, parms->pool, "add_log_context filename = [%s], linkname = [%s], env_condition = [%s], format_candidate = [%s], log_type = [0x%0x]", filename, linkname, env_condition, format_candidate, log_type);

    /* */
    sc = ap_get_module_config(parms->server->module_config, &log_access_module);

    /* allocate */
    if ( LOG_IS_ERROR_LOG(  log_type ) ) {
        lc = (log_context_t *) apr_array_push(sc->error_log_contexts);
    } else {
        lc = (log_context_t *) apr_array_push(sc->log_contexts);
    }

    /* */
    lc->log_type        = log_type;                                 /* */
    lc->filename         = apr_pstrdup(parms->pool, filename);        /* filename or pipename in strftime(3) format */    
    lc->linkname        = apr_pstrdup(parms->pool, linkname);       /* symbolic link name */    
    lc->env_condition   = apr_pstrdup(parms->pool, env_condition);  /* */

    if ( format_candidate == NULL ) {
        lc->log_format_string   = sc->default_format_string;
        lc->log_format_items    = format_items_from_format_string(parms->pool, sc->default_format_string, &error);

    } else if ((log_format_string = apr_table_get(log_format_table, format_candidate)) == NULL) {
        lc->log_format_string   = apr_pstrdup(parms->pool, format_candidate);
        lc->log_format_items    = format_items_from_format_string(parms->pool, format_candidate, &error);

    } else {
        lc->log_format_string = apr_pstrdup(parms->pool, log_format_string);
        lc->log_format_items  = format_items_from_format_string(parms->pool, lc->log_format_string, &error);
    }

    lc->file_interval   = log_timeutils_determine_interval(lc->filename);    /* */
    lc->link_interval   = log_timeutils_determine_interval(lc->linkname);   /* */

    if ( lc->link_interval < lc->file_interval ) {
        return "symbolic link filename may not roll more often than linked-to filename";
    }

    /* */
    return NULL;
}


/*
 * directive_transfer_log
 *
 * @directive   TransferLog
 *
 * @param       cmd_parms*
 * @param       void*
 * @param       const char*
 */

static const char *directive_transfer_log(cmd_parms *parms, void *unused, int argc, char *const argv[]) {

    server_context_t    *sc = NULL;
    const char          *error = NULL;

    sc = ap_get_module_config(parms->server->module_config, &log_access_module);

    switch ( argc ) {
        case 1:
            return add_log_context(parms, argv[0], NULL, NULL, NULL, 0);
        case 2:
            return add_log_context(parms, argv[0], argv[1], NULL, NULL, 0);
        default:
            error = "syntax: TransferLog <FILENAME> [<LINKNAME>]";
            break;
    }

    return error;
}


/*
 * directive_cookie_log
 *
 * @directive   CookieLog
 *
 * @param       cmd_parms*
 * @param       void*
 * @param       const char*
 */
static const char *directive_cookie_log(cmd_parms *parms, void *unused, const char *filename) {
    return add_log_context(parms, filename, NULL, NULL, "%{Cookie}n \"%r\" %t", 0);
}


/*
 * directive_clf_timestamps
 *
 * @directive   ClfTimestamps
 *
 * @param       cmd_parms*
 * @param       void*
 * @param       int
 */
static const char *directive_clf_timestamps(cmd_parms *parms, void *unused, int flag) {

    server_context_t *sc = NULL;

    sc = ap_get_module_config(parms->server->module_config, &log_access_module);
    
    if ( flag ) {
        ap_log_perror(APLOG_MARK, APLOG_DEBUG, 0, NULL, "CLF timestamps enabled");
    }

    sc->clf_timestamps = flag;

    return NULL;
}


/* ---- end directive processing callbacks ---- */


/* ---- begin APR static data ---- */

/* */
static const command_rec command_table[] = {

    /* mod_log_access */
    AP_INIT_TAKE_ARGV(  "AccessLog",        directive_access_log,       NULL, RSRC_CONF, "add access log" ),
    AP_INIT_TAKE_ARGV(  "LogFormat",        directive_log_format,       NULL, RSRC_CONF, "define log format, or set default log format" ),
    AP_INIT_FLAG(       "ClfTimestamps",    directive_clf_timestamps,   NULL, RSRC_CONF, "enable precise timestamping as required by CLF" ),

    /* deprecated, for compatibility with mod_log_config */
    AP_INIT_TAKE_ARGV(  "TransferLog",      directive_transfer_log,     NULL, RSRC_CONF, "set transfer log filename" ),
    AP_INIT_TAKE1(      "CookieLog",        directive_cookie_log,       NULL, RSRC_CONF, "set cookie log filename" ),

};

/* */
module AP_MODULE_DECLARE_DATA log_access_module = {

    STANDARD20_MODULE_STUFF,    /* STANDARD20_MODULE_STUFF      */
    NULL,                       /* init per-directory config    */
    NULL,                       /* merge per-directory config   */
    init_server_config,         /* init server config           */
    merge_server_config,        /* merge server config          */
    command_table,              /* const command_rec[]          */
    register_hooks              /* register hooks               */

};

/* ---- end APR static data ---- */


/*

---- directives ------------------------------------------------------

This module implements the TransferLog directive (same as the common log
module), and additional directives, LogFormat and AccessLog.

If no TransferLog is given, no logging is done.


---- syntax ----------------------------------------------------------

LogFormat <FORMAT>
LogFormat <FORMAT> <FORMAT_NAME>
    set default format for log files.
    define a log format called FORMAT called FORMAT_NAME

AccessLog <FILENAME> [link=<LINKNAME>] [format=<FORMAT>|<FORMAT_NAME>] [<CONDITION>]
    Log to file FILENAME with format FORMAT.
    FILENAME may be in strftime(3) format.
    FILENAME may start with a '|' character.
    LINKNAME is a symbolic link to the latest incarnation of FILENAME.
    LINKNAME may not contain strftime(3) escape characters.
    If CONDITION is set, events are only logged if CONDITION evaluates to  true.
    CONDITION is recognised as such by the presence of an '=' character.

TransferLog <FILENAME>
TransferLog <LINKNAME> <FILENAME>
    Logs transfers to FILENAME in standard log format, unless a custom format
    is set with LogFormat.  FILENAME may be in strftime(3) format, though this
    is not backwards compatible.  If set, LINKNAME is a symbolic link to the
    latest incarnation of FILENAME.  LINKNAME may not contain strftime(3)
    escape characters.

CookieLog <FILENAME>
    For backwards compatability with old Cookie logging module -- deprecated.

There may be any number of TransferLog and AccessLog commands.  Each request is
logged to ALL the named files, in the appropriate format.

If no TransferLog or AccessLog directives appear in a VirtualHost, the request
is logged to the log file(s) defined outside the VirtualHost section.

If a TransferLog or AccessLog directive appears in a VirtualHost section, the
log files defined outside the VirtualHost are not used.

This makes this module compatible with the CLF and config log modules, where
the use of TransferLog inside the VirtualHost section overrides its use
outside.


---- examples --------------------------------------------------------

    AccessLog   log/access_log

    <VirtualHost>
        LogFormat   "...custom format..."
        TransferLog log/virtual_only
        AccessLog   "log/agent" "%t %{user-agent}i"
        AccessLog   "log/%Y%m%d/%Y%m%d-ua-access.log" "%t %{user-agent}i"
        AccessLog   "log/%Y%m%d/%Y%m%d-ua-access.log" "log/today-ua-access.log"
        AccessLog   "log/%Y%m%d/%Y%m%d-ua-access.log" "log/today-ua-access.log" "%t %{user-agent}i"
    </VirtualHost>

Requests to the main server are logged to "log/access_log" in CLF.
Requests to the virtual host are logged to "log/virtual_only" using the format
specified by "...custom format..." and to "log/agent" using the custom
user-agent format. Log formats are defined by LogFormat


NCSA-format referer and user-agent logs are easily added with AccessLog:
    AccessLog   logs/referer  "%{referer}i -> %U"
    AccessLog   logs/agent    "%{user-agent}i"


RefererIgnore functionality can be obtained with conditional logging (SetEnvIf
and AccessLog ... env=!VAR), but using this method allows much easier
modification of the log format, e.g. to log hosts along with user-agent:
    AccessLog   logs/referer "%{referer}i %U %h"


---- '%' directives for LogFormat and AccessLog ----------------------

  %...A:           local IP address
  %...a:           remote IP address
  %...B:           bytes sent, excluding HTTP header fields
  %...b:           bytes sent, excluding HTTP header fields in CLF format (CLF convention is to use '-' instead of '0')
  %...{COOKIE}C:   value of the HTTP cookie named COOKIE
  %...D:           time taken to serve the request, in microseconds
  %...{ENVVAR}e:   value of the environment variable ENVVAR
  %...f:           filename
  %...h:           remote host
  %...H:           request protocol
  %...{HEADER}i:   contents of HEADER: header from request
  %...l:           remote user name (from identd, if supplied)
  %...m:           request method
  %...{FOOBAR}n:   contents of note FOOBAR from another module
  %...{HEADER}o:   contents of HEADER: header from reply
  %...P:           process ID of the child that serviced the request
  %...p:           port the request was served to
  %...{FORMAT}P:   process ID or thread ID of the child/thread that serviced the request
  %...q:           query string prepended by "?", or empty if no query string
  %...r:           first line of request
  %...s:           request status.  for internally redirected requests, status of the original request (use %...>s for the final request)
  %...T:           time taken to serve the request, in seconds
  %...t:           time in CLF time format
  %...{FORMAT}t:   time in strftime(3) format given by FORMAT
  %...U:           URL path requested
  %...u:           remote user (from auth; may be bogus if return status (%s) is 401)
  %...V:           server name according to the UseCanonicalName setting
  %...v:           configured name of the server (i.e. which virtual host?)
  %...X:           connection status
                     'X' = connection is aborted before the response completed
                     '+' = connection may be kept alive after the response is sent
                     '-' = connection is closed after the response is sent
                     (This directive was %...c in late versions of Apache 1.3, but this conflicted with the historical SSL %...{var}c syntax.)


The '...' can be anything at all (e.g. "%h %u %r %s %b"), or item can indicate
conditions for inclusion of the item (which causes item to be replaced with '-'
if the condition is not met).  Note that no escaping is performed on the
strings from %r, %...i and %...o;

some with long memories may remember that I thought this was a bad idea, once
upon a time, and I'm still not comfortable with item, but item is difficult to
see how to "do the right thing" with all of '%..i', unless we URL-escape
everything and break with CLF.

The forms of condition are a list of HTTP status codes, which may or may not be
preceded by '!'.  Thus '%400,501{User-agent}i' only logs 'User-agent:' on 400
errors and 501 errors (Bad Request, Not Implemented), '%!200,304,302{Referer}i'
logs 'Referer:' on all requests which did *not* return some sort of normal
status.

The default LogFormat reproduces CLF; see below.

The way this is supposed to work with virtual hosts is as follows:

a virtual host can have its own LogFormat, or its own TransferLog.
If item doesn't have its own LogFormat, item inherits from the main server.
If item doesn't have its own TransferLog, item writes to the same descriptor (meaning the same process for "| ...").

*/


/*
 * ---- credits ---------------------------------------------------------
 *
 *  who             email                       what
 *  --------------  --------------------------  -------------------------
 *  Andrew Ford     <a.ford@ford-mason.co.uk>
 *  David MacKenzie <djm@va.pubnix.com>
 *  Roger Beeman    <beeman@cisco.com>
 *
 */


/* 
 * ---- copyright -------------------------------------------------------
 *
 * Copyright 1999-2005 The Apache Software Foundation or its licensors, as
 * applicable.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
